/*!
 * Copyright (C) 2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

#include <cfg/os.h>
#include <cfg/clock.h>
#include <dev/board.h>
#include <dev/irqreg.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <sys/atom.h>
#include <sys/event.h>
#include <sys/timer.h>

#include "at73dac.h"

/*
 * Initial volume.
 */
#ifndef AT73DAC_VOL
#define AT73DAC_VOL 0x18
#endif

/*
 * Number of PCM buffers.
 */
#ifndef SAMPLE_BUFFERS
#if defined (AT91SAM9260_EK)
#define SAMPLE_BUFFERS  32
#else
#define SAMPLE_BUFFERS  3
#endif
#endif

/*
 * Select SPI interface.
 */
#ifndef DACSPI_ID
#if defined(AT91SAM7X_EK)       /* Board */
#define DACSPI_ID   SPI0_ID
#define DACSPI_PCS  SPI_PCS_2
#elif defined(AT91SAM9260_EK)   /* Board */
#define DACSPI_ID   SPI1_ID
#define DACSPI_PCS  SPI_PCS_0
#else                           /* Board */
#warning "Unknown board configuration"
#endif                          /* Board */
#endif

/*
 * Select SPI and I2S pins.
 */
#if defined(MCU_AT91SAM7X256)   /* MCU_ */

#define DACSPI_PIO_ID    PIOA_ID
#if DACSPI_ID == SPI0_ID
#define DACSPI_PINS_A  _BV(PA16_SPI0_MISO_A) | _BV(PA17_SPI0_MOSI_A) | _BV(PA18_SPI0_SPCK_A) | _BV(PA14_SPI0_NPCS2_A)
#define DACSPI_PINS_B  0
#elif DACSPI_ID == SPI1_ID
#define DACSPI_PINS_A  _BV(PA24_SPI1_MISO_B) | _BV(PA23_SPI1_MOSI_B) | _BV(PA22_SPI1_SPCK_B) | _BV(PA25_SPI1_NPCS1_B)
#define DACSPI_PINS_B  0
#endif

#define DACI2S_PIO_ID   PIOA_ID
#define DACI2S_PINS_A   _BV(PA23_TD_A) | _BV(PA21_TF_A) | _BV(PA22_TK_A)
#define DACI2S_PINS_B   0

#elif defined (MCU_AT91SAM9260) /* MCU_ */

#if DACSPI_ID == SPI0_ID
#define DACSPI_PIO_ID    PIOA_ID
#define DACSPI_PINS_A   _BV(PA0_SPI0_MISO_A) | _BV(PA1_SPI0_MOSI_A) | _BV(PA2_SPI0_SPCK_A) | _BV(PA3_SPI0_NPCS0_A)
#define DACSPI_PINS_B   0
#elif DACSPI_ID == SPI1_ID
#define DACSPI_PIO_ID    PIOB_ID
#define DACSPI_PINS_A   _BV(PB0_SPI1_MISO_A) | _BV(PB1_SPI1_MOSI_A) | _BV(PB2_SPI1_SPCK_A) | _BV(PB3_SPI1_NPCS0_A)
#define DACSPI_PINS_B   0
#endif

#define DACI2S_PIO_ID   PIOB_ID
#define DACI2S_PINS_A   _BV(PB18_TD0_A) | _BV(PB17_TF0_A) | _BV(PB16_TK0_A)
#define DACI2S_PINS_B   0

#else                           /* MCU_ */
#warning "Unknown MCU"
#endif                          /* MCU_ */

/*
 * Determine SPI registers.
 */
#if DACSPI_ID == SPI0_ID        /* DACSPI_ID */
#define DACSPI_MR   SPI0_MR
#define DACSPI_CR   SPI0_CR
#define DACSPI_SR   SPI0_SR
#define DACSPI_TDR  SPI0_TDR
#define DACSPI_RDR  SPI0_RDR
#if DACSPI_PCS == SPI_PCS_0
#define DACSPI_CSR  SPI0_CSR0
#elif DACSPI_PCS == SPI_PCS_1
#define DACSPI_CSR  SPI0_CSR1
#elif DACSPI_PCS == SPI_PCS_2
#define DACSPI_CSR  SPI0_CSR2
#elif DACSPI_PCS == SPI_PCS_3
#define DACSPI_CSR  SPI0_CSR3
#endif
#elif DACSPI_ID == SPI1_ID      /* DACSPI_ID */
#define DACSPI_MR   SPI1_MR
#define DACSPI_CR   SPI1_CR
#define DACSPI_SR   SPI1_SR
#define DACSPI_TDR  SPI1_TDR
#define DACSPI_RDR  SPI1_RDR
#if DACSPI_PCS == SPI_PCS_0
#define DACSPI_CSR  SPI1_CSR0
#elif DACSPI_PCS == SPI_PCS_1
#define DACSPI_CSR  SPI1_CSR1
#elif DACSPI_PCS == SPI_PCS_2
#define DACSPI_CSR  SPI1_CSR2
#elif DACSPI_PCS == SPI_PCS_3
#define DACSPI_CSR  SPI1_CSR3
#endif
#endif                          /* DACSPI_ID */

/*
 * Determine PIO used by SPI.
 */
#if DACSPI_PIO_ID == PIOA_ID    /* DACSPI_PIO_ID */
#define DACSPI_PDR  PIOA_PDR
#define DACSPI_ASR  PIOA_ASR
#define DACSPI_BSR  PIOA_BSR
#elif DACSPI_PIO_ID == PIOB_ID  /* DACSPI_PIO_ID */
#define DACSPI_PDR  PIOB_PDR
#define DACSPI_ASR  PIOB_ASR
#define DACSPI_BSR  PIOB_BSR
#endif                          /* DACSPI_PIO_ID */

/*
 * Determine PIO used by I2S.
 */
#if DACI2S_PIO_ID == PIOA_ID    /* DACI2S_PIO_ID */
#define DACI2S_PDR  PIOA_PDR
#define DACI2S_ASR  PIOA_ASR
#define DACI2S_BSR  PIOA_BSR
#elif DACI2S_PIO_ID == PIOB_ID  /* DACI2S_PIO_ID */
#define DACI2S_PDR  PIOB_PDR
#define DACI2S_ASR  PIOB_ASR
#define DACI2S_BSR  PIOB_BSR
#endif                          /* DACI2S_PIO_ID */

/*!
 * \brief I2S event queue.
 */
static HANDLE i2s_que;

/*!
 * \brief Number of channels.
 */
#define PCM_CHANS   2

/*!
 * \brief Number of bits per sample.
 */
#define PCM_BITS    16

/*!
 * \brief PCM buffer type.
 */
typedef struct _PCM_BUFFER {
    u_short *wbf_dat;   /*!< Pointer to PCM data. */
    int wbf_siz;        /*!< Buffer size. */
    int wbf_len;        /*!< Number of valid bytes buffered. */
} PCM_BUFFER;

/* PCM buffer queue. */
static PCM_BUFFER pcm_bufq[SAMPLE_BUFFERS];
/* PCM buffer queue. */
static volatile u_int brd_idx;
/* PCM buffer queue. */
static u_int bwr_idx;

/*!
 * \brief Set PDC pointer and counter registers.
 *
 * SSC interrupts must have been disabled befor calling this function.
 */
static void I2sPdcFill(void)
{
    if (brd_idx != bwr_idx) {
        if (inr(SSC_TNCR) == 0) {
            if (++brd_idx >= SAMPLE_BUFFERS) {
                brd_idx = 0;
            }
            outr(SSC_TNPR, (u_int) pcm_bufq[brd_idx].wbf_dat);
            outr(SSC_TNCR, pcm_bufq[brd_idx].wbf_len);
        }
    } 
}

/*!
 * \brief SSC interrupt handler.
 *
 * Called when PDC transmission done.
 */
static void I2sInterrupt(void *arg)
{
    I2sPdcFill();
    NutEventPostFromIrq(&i2s_que);
}

/*!
 * \brief Read value from specified AT73C213 register.
 *
 * Communicates with the DAC chip via SPI. Only a few bytes need to be 
 * transferred, thus we use polling.
 *
 * \param reg DAC register address.
 *
 * \return Register contents.
 */
u_char At73DacReadReg(u_int reg)
{
    u_int rc;

    /* Make sure that RDRF is cleared. */
    rc = inr(DACSPI_RDR);
    /* Set MSB to 1 for read operation, followed by the 7 bit register address. */
    outr(DACSPI_TDR, 0x80FF | (reg << 8));
    /* Wait until response is available in the receive data register. */
    while ((inr(DACSPI_SR) & SPI_RDRF) == 0);
    /* Read the result from the receive data register. */
    rc = inr(DACSPI_RDR);

    return (u_char) rc;
}

/*!
 * \brief Write value to specified AT73C213 register.
 *
 * Communicates with the DAC chip via SPI. Only a few bytes need to be 
 * transferred, thus we use polling.
 *
 * \param reg DAC register address.
 * \param val Value to store in specified register.
 */
void At73DacWriteReg(u_int reg, u_int val)
{
    u_int rc;

    /* Make sure that RDRF is cleared. */
    rc = inr(DACSPI_RDR);
    /* MSB is 0 for write operation, followed by the 7 bit register address. */
    outr(DACSPI_TDR, (reg << 8) | val);
    /* Wait until response is available in the receive data register. */
    while ((inr(DACSPI_SR) & SPI_RDRF) == 0);
    /* Read the result from the receive data register. Clears RDRF. */
    rc = inr(DACSPI_RDR);
}

/*!
 * \brief Configure the SPI rate.
 *
 * \param rate Baudrate. The maximum is MCK/1 and the minimum is MCK/255.
 *             If the specified rate is above the maximum or below the
 *             minimum, the maximum or minimum value resp. will be set.
 *
 * \return Always 0.
 */
static int At73SpiSetRate(u_long rate)
{
    u_int divider;

    /* The SPI clock is driven by the master clock. */
    divider = (u_int) At91GetMasterClock();
    /* Calculate the SPI clock divider. Avoid rounding errors. */
    divider += (u_int) (rate / 2);
    divider /= rate;
    /* A divider value of 0 is not allowed. */
    if (divider < 1) {
        divider = 1;
    }
    /* The divider value maximum is 255. */
    else if (divider > 255) {
        divider = 255;
    }
    outr(DACSPI_CSR, (inr(DACSPI_CSR) & ~SPI_SCBR) | (divider << SPI_SCBR_LSB));

    return 0;
}

/*!
 * \brief Configure SPI master to access the AT73C213.
 */
static void At73SpiInit(void)
{
    /* Enable peripherals for SPI. */
    outr(DACSPI_ASR, DACSPI_PINS_A);
    outr(DACSPI_BSR, DACSPI_PINS_B);

    /* Disable PIO lines used for SPI. */
    outr(DACSPI_PDR, DACSPI_PINS_A | DACSPI_PINS_B);

    /* Enable SPI clock. */
    outr(PMC_PCER, _BV(DACSPI_ID));

    /* Reset SPI. */
    outr(DACSPI_CR, SPI_SWRST);

    /* Set SPI to master mode, fixed peripheral at NPCS0, fault detection disabled. */
    outr(DACSPI_MR, (90 < SPI_DLYBCS_LSB) | DACSPI_PCS | SPI_MODFDIS | SPI_MSTR);

    /* 16 bit transfer. Data changes during clock low and will be sampled on rising edges. */
    outr(DACSPI_CSR, SPI_BITS_16 | (255 << SPI_SCBR_LSB));
    /* Min. SPI clock cycle time of the AT73C213 is 100ns. */
    At73SpiSetRate(10000);

    /* Enable SPI. */
    outr(DACSPI_CR, SPI_SPIEN);
}

/*!
 * \brief Disable I2S interface.
 *
 * \return Always 0.
 */
static int At73I2sDisable(void)
{
    /* Disable all interrupts. */
    outr(SSC_IDR, 0xFFFFFFFF);

    /* Disable SSC interrupt. */
    NutIrqDisable(&sig_SSC);

    /* Disable SSC clock. */
    outr(PMC_PCDR, _BV(SSC_ID));

    /* Reset receiver and transmitter. */
    outr(SSC_CR, SSC_SWRST | SSC_RXDIS | SSC_TXDIS);
    outr(SSC_RCMR, 0);
    outr(SSC_RFMR, 0);
    outr(SSC_PTCR, PDC_RXTDIS);
    outr(SSC_PTCR, PDC_TXTDIS);
    outr(SSC_TNCR, 0);
    outr(SSC_TCR, 0);

    return 0;
}

/*!
 * \brief Enable I2S interface with a given rate.
 *
 * For the AT73C213 it is required, that the falling edge of BCLK (our 
 * SSC transmit clock) follows the rising edge of MCLK (our PCK0 output) 
 * within a range of 2.5 to 7.5 ns. Thus we can't use the internal SSC 
 * clock divider, because we are not able to control the clock phase 
 * in relation to PCK0. In general BCLK equals MCLK divided by 32.
 *
 * On the SAM9260 we use timer/counter 1 to divide the clock at PCK0
 * by 32. This requires to connect PC1 (PCK0 output) to PB6 (TCLK1 input)
 * and PC7 (TIOB1 output) to PB16 (SSC0 transmit clock).
 *
 * \param rate PCM sample rate in Hertz.
 *
 * \return Always 0.
 */
static int At73I2sEnable(u_int rate)
{
    /* Enable clocks of timer 1 and SSC. */
    outr(PMC_PCER, _BV(TC1_ID) | _BV(SSC_ID));

    outr(TC1_CMR, TC_WAVE | /* PWM mode. */
        TC_CLKI | 
        TC_CLKS_XC1 | 
        TC_BCPB_SET_OUTPUT | 
        TC_BCPC_CLEAR_OUTPUT | 
        TC_EEVT_XC0 | 
        TC_CPCTRG);

    outr(TC1_RB, 4);
    outr(TC1_RC, 8);

    /* Enable TC1 output function at pin PC7. */
    outr(PIOC_ASR, _BV(PC7_TIOB1_A));
    outr(PIOC_PDR, _BV(PC7_TIOB1_A));

    /* Enable TCLK1 input function at pin PB6. */
    outr(PIOB_BSR, _BV(PB6_TCLK1_B));
    outr(PIOB_PDR, _BV(PB6_TCLK1_B));

    /* Enable the counter. */
    outr(TC1_CCR, TC_CLKEN);
    /* Software trigger starts the clock. */
    outr(TC1_CCR, TC_SWTRG);


    /* Select SSC peripheral functions. */
    outr(DACI2S_ASR, DACI2S_PINS_A);
    outr(DACI2S_BSR, DACI2S_PINS_B);

    /* Enable SSC peripheral pins. */
    outr(DACI2S_PDR, DACI2S_PINS_A | DACI2S_PINS_B);

    /* Configure 16-bit stereo I2S transmit format. */
    outr(SSC_TCMR,              /* Set transmit clock mode. */
         SSC_CKS_TK |          /* Use external clock. */
         SSC_START_FALL_RF |    /* Start transmission on falling edge at TF. */
         (1 << SSC_STTDLY_LSB) |        /* Delay start by 1 cycle. */
         ((((PCM_CHANS * PCM_BITS) / 2) - 1) << SSC_PERIOD_LSB));       /* New frame sync after 2 * (15 + 1) clocks. */
    outr(SSC_TFMR,              /* Set transmit frame mode. */
         ((PCM_BITS - 1) << SSC_DATLEN_LSB) |   /* Transmit 16 bits (1 half word). */
         SSC_MSBF |             /* Most significant bit first. */
         ((PCM_CHANS - 1) << SSC_DATNB_LSB) |   /* Transmit 2 data words per frame. */
         ((PCM_BITS - 1) << SSC_FSLEN_LSB) |    /* Length of frame sync. */
         SSC_FSOS_NEGATIVE);    /* Negative frame sync pulse. */

    /* Enable transmitter in PDC mode. */
    outr(SSC_PTCR, PDC_TXTEN);
    outr(SSC_CR, SSC_TXEN);

    return 0;
}

/*!
 * \brief Configure the SSC for I2S mode.
 *
 * \param rate Sampling rate in Hertz.
 *
 * \return Always 0.
 */
static int At73I2sInit(u_int rate)
{
    /* Register SSC interrupt handler and enable SSC interrupt. */
    NutRegisterIrqHandler(&sig_SSC, I2sInterrupt, 0);

    At73I2sDisable();
    At73I2sEnable(rate);
    NutIrqEnable(&sig_SSC);

    return 0;
}

/*!
 * \brief Initialize AT73C213 DAC interface.
 *
 * \param rate Sample rate.
 *
 * \return 0 on success, -1 otherwise.
 */
int At73DacInit(u_int rate)
{
    u_int pck_pres; /* PCK prescaler. */

    /* MCLK = 256 * Fs */
    pck_pres = At91GetMasterClock() / rate / 256;
    //pck_pres *= 2;
    printf("MCK=%lu rate=%u pres=%u\n", At91GetMasterClock(), rate, pck_pres);
    switch(pck_pres) {
    case 1:
        pck_pres = PMC_PRES_CLK;
        break;
    case 2:
        pck_pres = PMC_PRES_CLK_2;
        break;
    case 4:
        pck_pres = PMC_PRES_CLK_4;
        break;
    case 8:
        pck_pres = PMC_PRES_CLK_8;
        break;
    case 16:
        pck_pres = PMC_PRES_CLK_16;
        break;
    case 32:
        pck_pres = PMC_PRES_CLK_32;
        break;
    case 64:
        pck_pres = PMC_PRES_CLK_64;
        break;
    default:
        return -1;
    }

    /* Enable PCK0 output function at pin PC1. */
    outr(PIOC_BSR, _BV(PC1_PCK0_B));
    outr(PIOC_PDR, _BV(PC1_PCK0_B));

    /* Feed PCK0 from PLLA divided by the calculated prescaler. */
    outr(PMC_PCKR0, PMC_CSS_PLLA_CLK | pck_pres);

    /* Enable PCK0 output. */
    outr(PMC_SCER, PMC_PCK0);

    /* Initialize SPI. */
    At73SpiInit();

    /* Power down all amplifiers.*/
    At73DacWriteReg(DAC_CTRL, 0);

    /* Activate codec and filter reset. */
    At73DacWriteReg(DAC_RST, 0);
    /* De-activate codec and filter reset. */
    At73DacWriteReg(DAC_RST, DAC_RST_RSTZ | DAC_RST_RESFILZ);
    /* Mute. */
    At73DacWriteReg(DAC_LMPG, 0x20);
    At73DacWriteReg(DAC_RMPG, 0x20);
    //At73DacWriteReg(DAC_LLOG, 0x20);
    //At73DacWriteReg(DAC_RLOG, 0x20);
    /* Precharge and master power on. */
    At73DacWriteReg(DAC_PRECH, 0xFF);
    /* Power up DAC channels. */
    At73DacWriteReg(DAC_CTRL, DAC_CTRL_ONLNOL | DAC_CTRL_ONLNOR);
    NutSleep(500);
    /* Stop precharging, keep master power on. */
    At73DacWriteReg(DAC_PRECH, DAC_PRECH_ONMSTR);
    NutSleep(1);
    /* Power up line out drivers. */
    At73DacWriteReg(DAC_CTRL, DAC_CTRL_ONDACL | DAC_CTRL_ONDACR | DAC_CTRL_ONLNOL | DAC_CTRL_ONLNOR);

    /* Set data mode to 16-bit I2S. */
    At73DacWriteReg(DAC_MISC, DAC_MISC_NBITS_16 | DAC_MISC_DINTSEL_I2S);

    /* Set default volume. */
    At73DacWriteReg(DAC_LMPG, AT73DAC_VOL);
    At73DacWriteReg(DAC_RMPG, AT73DAC_VOL);

    /* Mute line in. */
    At73DacWriteReg(DAC_LLIG, 0x11);
    At73DacWriteReg(DAC_RLIG, 0x11);

    //At73DacWriteReg(DAC_MC, 0x03);

    /* Initialize I2S. */
    return At73I2sInit(rate);
}

/*!
 * \brief Start transmitting audio samples.
 *
 * \return Always 0.
 */
static int At73DacStart(void)
{
    NutIrqDisable(&sig_SSC);
    I2sPdcFill();
    outr(SSC_IER, SSC_ENDTX);
    outr(SSC_CR, SSC_TXEN);
    NutIrqEnable(&sig_SSC);

    return 0;
}

/*!
 * \brief Wait until all buffered samples have been transmitted.
 *
 * \return Always 0.
 */
int At73DacFlush(void)
{
    while (bwr_idx != brd_idx) {
        At73DacStart();
        NutEventWait(&i2s_que, 500);
    }
    return 0;
}

/*!
 * \brief Add audio samples to the AT73C213 transmit queue.
 *
 * \param buf Points to PCM data.
 * \param len Number of samples.
 *
 * \return 0 on success or -1 if out of memory.
 */
int At73DacWrite(void *buf, int len)
{
    u_int idx;

    /* Move to the next buffer to write to. */
    idx = bwr_idx + 1;
    if (idx >= SAMPLE_BUFFERS) {
        idx = 0;
    }

    /* If all buffers are filled, wait for an event posted by the
       interrupt routine. */
    while (idx == brd_idx) {
        if (NutEventWait(&i2s_que, 100)) {
            At73DacStart();
        }
    }

    /*
     * Check, if the current buffer size is too small or not allocated.
     */
    if (pcm_bufq[idx].wbf_siz < len) {
        if (pcm_bufq[idx].wbf_siz) {
            free(pcm_bufq[idx].wbf_dat);
            pcm_bufq[idx].wbf_siz = 0;
        }
        pcm_bufq[idx].wbf_dat = malloc(len * 2);
        if (pcm_bufq[idx].wbf_dat == NULL) {
            /* Out of memory. */
            return -1;
        }
        pcm_bufq[idx].wbf_siz = len;
    }

    /*
     * At this point we got an available buffer with sufficient size.
     * Move the data to it, set the number of valid bytes and update
     * the write (producer) index.
     */
    memcpy(pcm_bufq[idx].wbf_dat, buf, len * 2);
    pcm_bufq[idx].wbf_len = len;
    bwr_idx = idx;

    return 0;
}

/*!
 * \brief Set volume.
 *
 * Sets the master playback gain. Range is +12..-35 dB.
 *
 * \param left  Left channel volume.
 * \param right Right channel volume.
 *
 * \return 0 on success, -1 otherwise.
 */
int At73DacSetVolume(int left, int right)
{
    u_int lval = 0x1F;
    u_int rval = 0x1F;

    /* Cut to limits. */
    if (left > 12) {
        left = 12;
    }
    if (left < -36) {
        left = -36;
    }
    if (right > 12) {
        right = 12;
    }
    if (right < -36) {
        right = -36;
    }
    lval -= (left + 36) * 2 / 3;
    rval -= (right + 36) * 2 / 3;

    At73DacWriteReg(DAC_LMPG, lval);
    At73DacWriteReg(DAC_RMPG, rval);

    return 0;
}

